/*--------------------------------------------------------------------------
**      WinConfig.c                 Twin           From:  Twin/windows
**
**              
**------------------------- < License Information > ------------------------
**
**      This file was originally a part of Willows TWIN.  Willows
**  TWIN was released under a Library GPL (LGPL).  This permits
**  redistribution of this source code, provided that the full
**  TWIN license is in effect, and provided that all modifications
**  to this source code are made publicly available.
**  Please refer to Willows software (www.willows.com) or
**  LICENSE for full information.
**  
**      Under Twine, this file is also protected by an LGPL.  Please
**  see LICENSE for full details on this license.
**  
**
**      Copyright 1997 Willows Software, Inc.
**------------------------ < File Content Description > --------------------
**
**  Module:	windows/WinConfig.c
**
**  Description:
**      
**
**  Functions defined:
**    
**------------------------- < Revision Information > -----------------------
** $Header: /twinecvs/src/windows/WinConfig.c,v 1.5 2000/01/28 07:57:10 robf Exp $
**
**      Full Revision history at bottom of file
**      
**--------------------------------------------------------------------------*/


/*************************************************************************
 *
 * Manage configuration parameters through a single configuration file.
 * 
 * These environment variables control parameters:
 *
 * TWINERC		location of configuration file if set
 * 			default, current working directory, home, willows
 * HOME			location of users home directory if set
 * 			default, value retrieved from pwnam
 * LOGNAME		user name if set
 *			default, use value from pwnam
 * DISPLAY		display server if set
 * 			default, value retrieved from twinrc file
 * PATH			path used to find and open files,
 *			default if not in TWINRC is to use PATH
 *			
 *************************************************************************/

/*----------------------------------------------------------------------------
**              E X T E R N A L     I N C L U D E S 
**--------------------------------------------------------------------------*/
#include <stdio.h>
#include <stdlib.h>
#include <string.h>

/*----------------------------------------------------------------------------
**              I N T E R N A L     I N C L U D E S
**--------------------------------------------------------------------------*/
#include "Driver.h"
#include "WinConfig.h"
#include "platform.h"
#include "compat.h"

/**************************************************************************
 *
 * [boot]		startup configuration settings
 * [windows]		run-time configuration settings
 * [xdos]		mapping of DOS drives
 * [commondialogs]	how to drive common dialog fileopen dialog
 *
 * [fontalias]		what font mapping to use
 * [colors]		what colors to use
 *  
 **************************************************************************
 *
 *	parameter	what parameter is being addressed
 *	opcode	
 *		1	use environment as override if set
 *			use section, keyname, default to lookup string
 *		6	use environment as override if set
 *			use section, keyname, default to lookup string
 *			if null, return path to TWINRC
 *
 *	lpszsection	section to use
 *	lpszkeyname	keyname to use
 *	lpszdefault	default value
 *	lpszenviron	environment string
 *	lpszstring	allocated buffer
 *
 *************************************************************************/


static char *sections[] = {
	"boot",
	"windows",
	"XDOS",
	"CommonDialogs"
};

static TWIN_CONFIG  *wcpdata = NULL;


/* name of configuration file */
extern BOOL TWIN_InitDriver();

BOOL InitTwinConfig();

char *
GetTwinFilename(void)
{
	static char *twinrc;

	/* have we already set it */
	if(twinrc)
		return twinrc;

	twinrc = (char *) DRVCALL_CONFIG(PCFGH_GETFILENAME, 0, 0 , NULL);

	return twinrc;
	
}


char *
GetTwinBuffer(TWIN_CONFIG *tp,char *wcpbuf)
{
	/* if we have a buffer, use it */
	if(wcpbuf)
		return wcpbuf;

	/* if we already allocated a buffer, return it */
	if(tp->lpszstring)
		return tp->lpszstring;

	/* allocate a permanent buffer */
	tp->lpszstring = (char *) WinMalloc(256);

	/* return it */
	return tp->lpszstring;	
}

char *
GetTwinString(int wcp, char *wcpbuf,int nsize )
{
	TWIN_CONFIG  *tp;
	char     *p = NULL;
	static char *twinrc;

	/* test if we have one, and load it with defaults */
	InitTwinConfig();

	if(twinrc == 0)
		twinrc = GetTwinFilename();
	
	if (wcpdata == NULL)
		wcpdata = (TWIN_CONFIG *)DRVCALL_CONFIG(PCFGH_GETDEFAULTS, 0, 0 , NULL);
	for(tp = wcpdata;tp->parameter > 0;tp++) {
		if(tp->parameter == wcp) {

		  /* see if we are to use an environment variable, and */
		  /* see if it is in fact set */

		  if(tp->lpszenviron && 
			   (p = (char *) getenv(tp->lpszenviron))) {
			       strcpy(GetTwinBuffer(tp,wcpbuf),p);

			       /* we got the environment variable */
			       return GetTwinBuffer(tp,wcpbuf);
  		  }

		  /* setup where we will store static info for this element */
		  /* p is where to put the answer */
		  p = GetTwinBuffer(tp,wcpbuf);
		  if(nsize == 0)
			nsize = 256;

		  /* return answer in p */
		  GetPrivateProfileString(
			sections[tp->section],
			tp->lpszkeyname,
			tp->lpszdefault,
			p,
			nsize, 
			twinrc);

		  /* return whatever we got, including NULL */
		  if(tp->opcode == 1) {
			/* answer is already in correct place */
			/* from above */
			return GetTwinBuffer(tp,wcpbuf);

		  }
		
		  /* if we got something return it */
		  if(strlen(p)) {
			return p;
		  }
				
		  /* we didnt get anything, so return a default */
		  strcpy(p, GetTwinFilename());

		  /* worst case same as TwinFilename directory */
		  return p;
		}
	}

	return 0;
}

typedef	struct {
	int	value;
	int     init;
	char   *string;
} CONFIGOPTION;

static CONFIGOPTION configoption[WD_LASTVALUE+1] = {
	{ 0, 0,"debugger"},	/* load the the debugger 		*/
	{ 0, 0,"noditherpure"},	/* how to handle dithering      	*/
	{ 0, 0,"colormap"},	/* allocate private colormap    	*/
	{ 0, 0,"hiresfonts"},	/* use hi-res fonts, not low 		*/
	{ 0, 0,"vgamode"}, 	/* default to 640x480x16colors  	*/
	{ 0, 0,"nothreads"},	/* don't allow threads			*/
	{ 0, 0,"fatalerr"},	/* break into debugger 			*/
	{ 0, 0,"handlerr"},	/* break on handleobj errors 		*/
	{ 0, 0,"noapistubs"},	/* NO msgbox for STUB 			*/
	{ 0, 0,"nopaging"},  	/* disable demand page loading  	*/
	{ 0, 0,"nochangecase"},	/* don't change case of filenames 	*/
	{ 0, 0,"mapcurdir"},	/* xdos map current directory   	*/
	{ 0, 0,"colormask"},	/* use rgb colormask to convert mono    */

	{ 0, 0,"failerr"},	/* break on log FAIL errors 		*/
	{ 0, 0,"warnerr"},	/* break on log WARN errors 		*/
	{ 0, 0,"fixmeerr"},	/* break on log FIXME errors 		*/
	{ 0, 0,"stuberr"},	/* break on log STUB  errors 		*/
	{ 0, 0,"syncmode"},	/* break on log STUB  errors 		*/
};	

int
GetConfigOption(int uid)
{
	char           *twinrc = GetTwinFilename();
        CONFIGOPTION   *cp     = 0;
	int    		rc     = 0;

	if(uid < 0 || uid >WD_LASTVALUE)
		return 0;

	cp = &configoption[uid]; 

#ifdef DEBUG
	if(uid == WD_DEBUGGER)
		cp->value = 1;
#endif
	
	if(cp) {
		if(cp->init)
			return cp->value;

		rc = GetPrivateProfileInt("boot",
			cp->string,	// keyname
			cp->value,	// default value
			twinrc);	

		cp->value = rc;
		cp->init  = 1;
	}

	return rc;
}

/********************************************************************
*	GetEnv
*
*	Our version of getenv that reads the [Environment] section of the twinrc file to obtain variables.
********************************************************************/
char *GetEnv(const char *env)
{
	char *twinrc = GetTwinFilename();
	static char buf[256];
	int count;

	/* Try and retrieve the environment variable */
	count = GetPrivateProfileString(sections[WCS_ENV], env, NULL, buf, 256,
		twinrc);

	if (count == 0)
		/* The default was used, meaning nothing was found */
		return(NULL);
	else
		/* Return the buffer containing the value found in the twinrc
		 * [Environment] section.
		 */
		return(buf);
}

BOOL InitTwinConfig()
{
    static BOOL binit = FALSE;

    if (!binit) {
     FILE*           fp;
     TWIN_CONFIG  *tp;
     char *twinrc;
     char buf[MAX_PATH];

     binit = TRUE;

     twinrc = GetTwinFilename();

     if (wcpdata == NULL)
          wcpdata = (TWIN_CONFIG *)DRVCALL_CONFIG(PCFGH_GETDEFAULTS, 0, 0 ,
NULL);

     /* Try to open .twinrc */
     fp = fopen(twinrc, "r");

     if (!fp) { /* Failed : create default with only boot section */
         fp = fopen(twinrc, "w");
         fprintf(fp,"%s\n","[boot]");

         for(tp = wcpdata;tp->parameter > 0;tp++) {
          if ( tp->parameter && tp->section == WCS_BOOT &&
               tp->lpszkeyname && *(tp->lpszkeyname) ) {
              strcpy(buf,tp->lpszkeyname);
              strcat(buf,"=");
              strcat(buf,tp->lpszdefault ? tp->lpszdefault : "");

              fprintf(fp,"%s\n",buf);
          }
         }
     }

     fclose(fp);
    } /* if (!binit) */

   return TRUE;
}

/*------------------------- < Full Revision History > ----------------------
** $Log: WinConfig.c,v $
** Revision 1.5  2000/01/28 07:57:10  robf
** simplify interface
**
** Revision 1.4  1999/07/09 21:55:45  mwalsh
** Updated comments to standard header and footer
**
**-------------------------------------------------------------------------*/
