/*--------------------------------------------------------------------------
**      utils.c                 Twin           From:  Twin/windows
**
**              
**------------------------- < License Information > ------------------------
**
**      This file was originally a part of Willows TWIN.  Willows
**  TWIN was released under a Library GPL (LGPL).  This permits
**  redistribution of this source code, provided that the full
**  TWIN license is in effect, and provided that all modifications
**  to this source code are made publicly available.
**  Please refer to Willows software (www.willows.com) or
**  LICENSE for full information.
**  
**      Under Twine, this file is also protected by an LGPL.  Please
**  see LICENSE for full details on this license.
**  
**
**      Copyright 1997 Willows Software, Inc.
**------------------------ < File Content Description > --------------------
**
**  Module:	windows/utils.c
**
**  Description:
**      
**
**  Functions defined:
**    
**------------------------- < Revision Information > -----------------------
** $Header: /twinecvs/src/windows/utils.c,v 1.14 2000/01/28 07:59:59 robf Exp $
**
**      Full Revision history at bottom of file
**      
**--------------------------------------------------------------------------*/

/***************************************************************************
 *
 * apiflag values
 *
 * LM_NOHEADER		1	no API CALL/API RET header
 * LM_NODEPTH		2	no nesting level
 * LM_APICOUNT		4	show api count
 * LM_APIINDENT		8	indent lines
 *
 * LM_BC_STYLE		0x10	suppress API CALL messages, ala BoundsChecker
 * LM_APISTARTUP	0x20	do initial logging before logstr initialized
 *
 ***************************************************************************/
/*----------------------------------------------------------------------------
**              E X T E R N A L     I N C L U D E S 
**--------------------------------------------------------------------------*/
#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>

/*----------------------------------------------------------------------------
**              I N T E R N A L     I N C L U D E S
**--------------------------------------------------------------------------*/
#include "windows.h"
#include "utils.h"
#include "Log.h"
#include "compat.h"
#include "ctype.h"
#include "debugdefs.h"
#include "Token.h"
 
static char format[4096];
static char title[256];
static char lineout[256];
static int apistate;
static int apistart;
static int apidepth;
static int apicount;
static LONG apithread;
static int  apiflag;
static int  apionoff;
static int  apilogging;

/* Initialization routine for library core */
extern void TWIN_InitDriver();

/*
 *	semantics:
 *	debug = [class]{+|-}channel[=stream]
 *	1	class is one of builtin 8
 *		       debug fixme fail	warn stub log call ret
 *
 *      2	channels are builtin
 *			API, LIB, X11, MSG, INT, DOS, BIN, SYS
 *		or any user defined
 *			ie. WARN(profile,".twinrc does not exist\n");
 *		or the special class all
 *	
 *	3	stream is one of
 *		0	no output, same as -channel
 *		1	stdout
 *		2	stderr
 *		3	logfile as defined by WIN_OPENLOG
 *
 *	[tools]
 *	debug = warn+api,trace-lib,-X11,+sys=2
 *
 *	debug class is left to right
 *
 */

typedef struct {
	int     init;
	int	fd;
	char   *file;
} TWINDEBUG;

static TWINDEBUG	TwinDebug;

void
DEBUG_mark(char *s)
{
	apilogging++;
	logstr(LF_CONSOLE,"*********** Mark %s **************\n",s);
}

void
DEBUG_apilogging(int flag)
{
	apilogging = flag;
}

static int
opnlog()
{
        char	*fn;
	char	*lf;
	char    *lc;

	if (TwinDebug.fd)
		return(TwinDebug.fd); 

	if (TwinDebug.file)
		fn = TwinDebug.file;
	else
		fn = (char *)getenv("WIN_OPENLOG");
	lf = (char *)getenv("WIN_APIFLAG");
	if ( lf ) {
		apiflag = strtol(lf,0,0);
		printf("apiflag=%x\n",apiflag);
		if(apiflag & LM_NOHEADER)
		   printf("	no API CALL/API RET header\n");
		if(apiflag & LM_NODEPTH)
		   printf("     no nesting level\n");
		if(apiflag & LM_APICOUNT)
		   printf("	show api count\n");
		if(apiflag & LM_APIINDENT)
		   printf("	indent lines\n");
		if(apiflag & LM_BC_STYLE)
		   printf("	suppress API CALL messages\n");
		if(apiflag & LM_APISTARTUP)
		   printf("	do initialization logging\n");
	}

	lc = getenv("WIN_APISTART");
	if ( lc ) {
		apistart = strtol(lc,0,0);
		printf("apistart=%d\n",apistart);
	}

	apilogging = 1;
	lc = getenv("WIN_APILOGGING");
	if ( lc ) {
		apilogging = strtol(lc,0,0);
		printf("apilogging=%d\n",apilogging);
	}

	if ( fn ) {
		char *p;

		if( *fn == '-') 
			return (TwinDebug.fd = 1);

		/* get rid of old style mask... */
		if ((p = strchr(fn,' ')))
			*p = 0;	
		return (TwinDebug.fd = creat(fn, 0666));
	}

	return 0;
}

void
DumpMemory(char *s,char *p,int n)
{
	int i,j;
	char x;
	char   *chrptr;
	char	buffer[132];
	char	temp[80];

	if(n==0)
		return;

	logstr(LF_CONSOLE,"%s:\n",s?s:"DumpMemory");
	if(p == 0) {
		logstr(LF_CONSOLE,"NULL pointer\n");
		return;
	}
	for(i=0;i<n;i++) {
		sprintf(buffer,"%x: ",(unsigned int) p);
		chrptr = p;

		for(j=0;j<16;j++) {
			x=*p++;
			x &= 0xff;
			sprintf(temp,"%2.2x ",x&0xff);
			strcat(buffer,temp);
		}		
		strcat(buffer,"   ");
		p = chrptr;
		chrptr = buffer + strlen(buffer);
		for(j=0;j<16;j++) {
			if(isprint(*p))
				*chrptr++ = *p;
			else	*chrptr++ = '.';
			p++;
		}	
		*chrptr = 0;

		i += 16;
		logstr(LF_CONSOLE,"%s\n",buffer);
	}
}

extern void debuggerbreak();

static int
logbreak(int reason)
{
	int dobreak = 0;

	switch(reason) {
	case dbcl_FIXME:
		dobreak = WD_FIXERR;
		break;
	case dbcl_WARN:
		dobreak = WD_WARNERR;
		break;
	case dbcl_STUB:
		dobreak = WD_STUBERR;
		break;
	case dbcl_FAIL:
	default:
		dobreak = WD_FAILERR;
		break;
		
	}
	if(dobreak && GetConfigOption(dobreak))
		debuggerbreak();
	return 0;
}
int
DEBUG_logonoff(int flag)
{
	int	cflag = apionoff;
	apionoff = flag;
	return cflag;
}

/*VARARGS */
void
logstr(unsigned long flg, ...)
{
	va_list args;
	int  fd = 0;
	char *fmt;
	char *header = 0;
	short class,channel;

	va_start(args, flg);

	/* point to fmt and args... */
	fmt = va_arg(args, char *);
	channel = flg>>8;
	class   = flg & 0xff;

	/* Should NOT happen */
	if(TwinDebug.init == 0) {
		TWIN_InitDriver();
	} 

	fd = debug_msg_enabled[channel][class];

	switch(fd) {
	case 0:
		break;
	case 1:
		break;
	case 2:
		break;
	case 3: 
		fd = TwinDebug.fd;
	}
	if(apistate == 0)
		fd = 0;		
	
	/* we may not be logging on this class/channel */
	if(fd) {

		*lineout = (char) 0;
		switch(class) {
			case dbcl_CONSOLE:
				break;

			case dbcl_FIXME:
			case dbcl_WARN:
			case dbcl_STUB:
				fd = 
				logbreak(class);
				/* print out a header */
				sprintf(title,"%s %s         ",
					(char *)debug_ch_name[channel],
					(char *)debug_cl_name[class]);
				sprintf(format,"%-14.14s : ",title);
				strcat(lineout,title);
				header = title;
				break;

			case dbcl_FAIL:
				/* check if we should break */
				logbreak(class);	

				/* fall through */
			case dbcl_TRACE:
			case dbcl_CALL:
			case dbcl_RET:
				if(channel != dbch_API)
					break;

				if(class == dbcl_RET || class == dbcl_FAIL)
					apidepth--;

				/* are we in same thread as last time? */
				if(apithread != GetCurrentThreadId()) {
					apithread = GetCurrentThreadId();
					sprintf(title,"\nThreadID %x\n",
					    (UINT)GetCurrentThreadId());
					strcat(lineout,title);
				}

				if((apiflag & LM_NOHEADER)  == 0) {
					/* print out a header */
					sprintf(title,"%-3.3s %-5.5s ",
						(char *)debug_ch_name[channel],
						(char *)debug_cl_name[class]);
					strcat(lineout,title);
				}
				if((apiflag & LM_NODEPTH)  == 0) {
					/* print out a apidepth */
					sprintf(title,"%2d ", apidepth);
					strcat(lineout,title);
				}
				if((apiflag & LM_APICOUNT)) {
					/* print out a apicount */
					sprintf(title,"%d ", apicount);
					strcat(lineout,title);
				}

				strcat(lineout,": ");

				if((apiflag & LM_APIINDENT) && apidepth > 0 && apidepth < 256) {
					memset(title,0,256);
					memset(title,' ',apidepth);
					strcat(lineout,title);
				}
				if(class == dbcl_CALL) {
					if(apiflag & LM_BC_STYLE)
						fd = 0;
					apicount++;
					apidepth++;
				}

				if(apicount < apistart)
					fd = 0;

				if(apilogging == 0)
					fd = 0;

				header = lineout;

				break;
		}
		if(fd) {
			if(header) {
				write(fd, lineout, strlen(lineout));
			}
			
			/* format the message */
			vsprintf(format,fmt,args);

			/* print out remainder of message locally */
			write(fd, format, strlen(format));

			/* if fd is NOT the log file, put it there anyway */
			if((fd != TwinDebug.fd) && TwinDebug.fd)
				write(TwinDebug.fd,format,strlen(format));
		}
	}

	va_end(args);
}

/* fixme+all=2 fail+all=2 warn+all=2 stub+all=3  */
/* debug+all=2 trace+all=3 call+all=3 ret+all=3  */
/* fixme+api=3 fail+api=3 warn+api=3 debug+api=3 */

void
DEBUG_TraceFlag(char *lpstr)
{
	int  cl,ch,fd;
	char *s,*p;
	char *class = "";

	fd = 3;
	s = strchr(lpstr,'-');
	if(s==0)
	    s = strchr(lpstr,'+');
	p = strchr(lpstr,'=');
	if(s && *s == '-') {
		fd = 0;
		*s++ = 0;
		class = lpstr;
		lpstr = s;
	}
	if(s && *s == '+') {
		fd = 3;
		*s++ = 0;
		class = lpstr;
		lpstr = s;
	}

	if(p) {
		*p++ = 0;
		/* TODO if a filename, open and pass its fd? */
		fd = atoi(p);
	}

	if(class && strlen(class) == 0) class = 0;
	if(lpstr && strcasecmp(lpstr,"all") == 0) lpstr = 0;

	for(ch=0;ch<DEBUG_CHANNEL_COUNT;ch++) {
	   if(lpstr == 0 || strcasecmp(lpstr,debug_ch_name[ch]) ==0) {
		for(cl=0;cl<8;cl++) {
			if(class == 0 || strcasecmp(class,debug_cl_name[cl]) == 0) {
				debug_msg_enabled[ch][cl] = fd;

			}
		}
		/* if we had a channel, we are done */
		if(lpstr)
			break;
            }
	}
}

extern char *GetTwinFilename();
void 
InitDebugSystem()
{
    	TOKEN tk;
	char trace[256];
    	char *p;

	if(TwinDebug.init == 0) {

		TwinDebug.init++;
		TwinDebug.fd = opnlog();
		if(apiflag & LM_APISTARTUP)
			apistate = 1;

    		GetPrivateProfileString((LPSTR)"boot",
			"trace",
			"",
			trace,
			256,
			GetTwinFilename());

    		if(strlen(trace)) {
			for(p = TWIN_FirstString(&tk,trace," \t,");p;p = TWIN_NextString(&tk))			{
				DEBUG_TraceFlag(p);
			}
    		} 
		
    		apiflag = GetPrivateProfileInt((LPSTR)"boot",
			"apiflag",
			apiflag,
			GetTwinFilename());
		apistate = 1;
	}	
}


#if 0
 	FIXME(LIB,"WindowsDirectory = %s\n",GetWindowsDirectory());
 	ERR(LIB,"WindowsDirectory = %s\n",GetWindowsDirectory());
	APICALL(API,"GetTickCount()\n");
	APIRET(API,"GetTickCount()\n");
	TRACE(X11,"This is a message\n");
	TRACE(MSG,"[HWND=%x MSG=%x WPARAM=%x LPARAM=%x]\n",h,m,w,l);

	FAIL(INT,"failed\n");
	FAIL(DOS,"failed\n");
	TRACE(SYS,"Driver call\n");
	STUB(API,"foo()\n");
	DEBUG(BIN,"PE Loader: find debug symbols\n");
	FIXME("WindowsDirectory = %s\n",GetWindowsDirectory());
	WARN(".twinrc does not exist\n");
#endif

/*------------------------- < Full Revision History > ----------------------
** $Log: utils.c,v $
** Revision 1.14  2000/01/28 07:59:59  robf
** added simple interface
**
** Revision 1.13  1999/12/09 07:52:52  robf
** removed unused lockfile
**
** Revision 1.12  1999/11/29 23:36:25  robf
** added logstr enhancements
**
** Revision 1.11  1999/11/29 07:15:30  robf
** logging update
**
** Revision 1.10  1999/10/25 18:10:11  jsheets
** Added DEBUG_EnableLogging() function to interactively turn logging on/off at runtime
**
** Revision 1.9  1999/07/31 02:09:20  jwhite
** Revised to add an ERR log macro, to force a scan to pick it up.
**
** Revision 1.8  1999/07/09 21:55:48  mwalsh
** Updated comments to standard header and footer
**
**-------------------------------------------------------------------------*/
